from bs4 import BeautifulSoup
import os
import config
import re
import numpy as np

from model import Mention, Entity


def from_isnotes(path):
    
    split_path = path.split(r"/")
    root, file_name = "/".join(split_path[:-1]), split_path[-1]
    
    mentions = []
    
    if file_name.endswith(".mmax"):
        
        #print(file_name)
        
        file_root = file_name[:-5]
        
        basedata = BeautifulSoup(open(root+"/Basedata/"+file_root+"_words.xml", "r").read(), features="lxml")
        entities = BeautifulSoup(open(root+"/markables/"+file_root+"_entity_level.xml", "r").read(), features="lxml")
        coref = BeautifulSoup(open(root+"/markables/"+file_root+"_coref_level.xml", "r").read(), features="lxml")
        
        for entity in entities.contents[2].contents[0].contents[0].find_all():
            
            span = entity["span"]
            tokens = []
            mention_id = entity["id"]
            
            # comparative: either "outsidetext" or antecedent id
            comp_from = None
            if entity["information_status"] == "mediated":
                try:
                    if entity["mediated_type"] == "comparative":
                        comp_from = entity["comparative_type"]
                        if comp_from == "withintext":
                            comp_from = entity["comp_from"]
                except:
                    pass
            
            try:
                coref_set = coref.find("markable", {"span": span})["coref_set"]
            except:
                coref_set = None
            
            if ".." in span:
                
                span_ids = [int(re.sub("word_", "", word)) for word in span.split("..")]
                for i in range(span_ids[0],span_ids[1]+1):
                    token = basedata.find("word", {"id": "word_"+str(i)}).contents[0]
                    tokens.append(token)
                
                mentions.append(Mention( tokens, [span_ids[0],span_ids[1]], mention_id, coref_set, comp_from))
                
            else:
                
                token = basedata.find("word", {"id": span}).contents[0]
                tokens.append(token)
                
                word_id = int(span.split("_")[-1])
                mentions.append(Mention(tokens, [word_id, word_id], mention_id, coref_set, comp_from))
    
    return mentions

def from_conll(path):
    #pass
    conll = open(path, 'r')
    
    mentions = list()
    
    text = list()
    poss = list()
    
    word_id = 0
    mention_id = 0
    coref_id = 0
    
    comp_count = 0
    ment_count = 0
    
    stack = list()
    coref_stack = dict()
    
    for line in conll:
        if line.startswith("#begin document"):
            pass
        elif line.startswith("#end document"):
            pass
        elif line == "\n":
            stack = list()
        else:
            naked_line = line.strip()
            split_line = naked_line.split()
            # Important indices:
            # 1:doc_id, 2:token_id,     3:token,
            # 4:POS,    5:syntax_tree, -1:Coref
        
            coref = split_line[-1].split("|")
            
            coref_stack = {**coref_stack, **{int(c): word_id for c in re.findall("\(([0-9]+)", split_line[-1])}}
            
            coref_spans = dict()
            for c in re.findall("([0-9]+)\)", split_line[-1]):
                if int(c) not in coref_stack:
                    raise Exception
                coref_spans[(coref_stack[int(c)], word_id)] = int(c)
            
            syntax = split_line[5].split("*")
            
            opening = [str(f) for f in re.findall("\(([A-Z]+)",syntax[0])]
            for paren in opening:
                stack.append((word_id, paren, split_line[3], split_line[4]))
            
            
            text.append(split_line[3])
            poss.append(split_line[4])
            
            closing = [str(f) for f in re.finditer("\)",syntax[1])]
            
            
            for paren in closing:
                start = stack[-1]
                if start[1] == "NP" or start[1] == "NML" or start[1] == "NX" or start[1] == "NAC":
                    ment_count += 1
                    
                    rules = list()
                    
                    rules.append(start[3]=="JJR")
                    
                    rules.append(text[start[0]].lower() == "more" 
                                 and len(poss) > start[0]+1 
                                 and poss[start[0]+1] == "JJ")
                    rules.append(text[start[0]].lower() in config.COMPARATIVES)
                    rules.append(text[start[0]].lower() == "others")
                    
                    if any(rules) and not "than" in text[start[0]:word_id+1]:
                        comp_count += 1
                        comp_from = "undefined"
                    else:
                        comp_from = None
                    
                    if (start[0],word_id) in coref_spans:
                        coref_set = coref_spans[(start[0],word_id)]
                    else:
                        coref_set = None
                    
                    mentions.append(Mention(text[start[0]:word_id+1], [start[0], word_id], mention_id, coref_set, comp_from))
                    
                stack = stack[:-1]
                
                
                
                
                mention_id += 1
            
            
            word_id += 1
            
    return mentions
            
            
def to_mmax(source_path, target_directory_path):
    """ Source must be in conll format
    
    """
    
    name = os.path.splitext(os.path.basename(source_path))
    base_name = name[0]
    
    print("\n[NCR] Extracting entities from \033[1m{}\033[0m.".format("".join(name)))
    
    mentions = from_conll(source_path)
    
    comp_from_list = [m.comp_from for m in mentions]
    
    len_mentions = sum(bool(c) for c in comp_from_list)
        
    
    new_name = "{:0>2}_{}".format(len_mentions, base_name)
    
    if any(comp_from_list):
        
        mmax = open(os.path.join(target_directory_path,  new_name+".mmax"), "w")
        mmax.write("""<?xml version="1.0"?>
<mmax_project>
<!--<sentences>002_htc_text.xml</sentences>-->
<turns></turns>
<words>{}</words>
<gestures></gestures>
<keyactions></keyactions>
</mmax_project>
""".format(new_name+"_words.xml"))
        mmax.close()
        
        target_path = os.path.join(target_directory_path, "Basedata", new_name+"_words.xml")
        
        source = open(source_path, 'r')
        target = open(target_path, 'w')
        
        word_id = 0
        
        
        target.write(r"""<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE words SYSTEM "words.dtd">
<words>""")
        
        for line in source:
            if line.startswith("#begin document"):
                pass
            elif line.startswith("#end document"):
                pass
            elif line == "\n":
                pass
            else:
                naked_line = line.strip()
                split_line = naked_line.split()
                
                token = split_line[3]
                if token == "&":
                    token = r"&amp;"
                token = re.sub("<", "&lt;", token)
                token = re.sub(">", "&gt;", token)
                
                target.write(r'<word id="word_'+str(word_id)+r'">'+token+r'</word>'+'\n')
                
                word_id += 1
        
        target.write(r"</words>")
        
        print("[NCR] Writing \033[92m{:>4}\033[0m token(s) to \033[1m{}_words.xml\033[0m.".format(word_id, new_name))
        
        markable_id = 0

        entity_level_path = os.path.join(target_directory_path+"Markables", new_name+"_entity_level.xml")
        entity_level = open(entity_level_path, 'w')
        entity_level.write(r"""<?xml version='1.0'?>
<!DOCTYPE markables SYSTEM "markables.dtd">
<markables xmlns="www.comp.leeds.ac.uk/markert/entity">""")
        
        coref_level_path = os.path.join(target_directory_path+"Markables", new_name+"_coref_level.xml")
        coref_level = open(coref_level_path, 'w')
        coref_level.write(r"""<?xml version='1.0'?>
<!DOCTYPE markables SYSTEM "markables.dtd">
<markables xmlns="www.eml.org/NameSpaces/coref">""")
        
        coref_chains = set()
        
        for mention in mentions:
            if mention.comp_from:
                entity_level.write(r'<markable id="markable_'+str(markable_id)+r'" information_status="mediated" mediated_type="comparative" comparative_type="withintext" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" mmax_level="entity" />'+'\n')
                markable_id += 1
            else:
                entity_level.write(r'<markable id="markable_'+str(markable_id)+r'" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" mmax_level="entity" />'+'\n')
                markable_id += 1
            
            if mention.coref_set:
                coref_chains.add(mention.coref_set)
                coref_level.write(r'<markable id="markable_'+str(markable_id)+r'" span="word_'+str(mention.span[0])+r'..word_'+str(mention.span[1])+r'" coref_set="set_'+str(mention.coref_set)+r'" mmax_level="coref" />'+'\n')
                markable_id += 1
            else:
                pass
        
        entity_level.write(r'</markables>')
        coref_level.write(r'</markables>')
        
        entity_level.close()
        coref_level.close()
    
        print("[NCR] Writing \033[92m{:>4}\033[0m mention(s) to \033[1m{}_entity_level.xml\033[0m.".format(len_mentions, new_name))
        print("[NCR] Writing \033[92m{:>4}\033[0m coreference chain(s) to \033[1m{}_coref_level.xml\033[0m.".format(len(coref_chains), new_name))
        
    else:
        print("[NCR] \033[1m{}\033[96m does not contain any comparatives.\033[0m".format(base_name))
        
    
if __name__ == "__main__":
    # ISnotes
    #root = config.DATA_PATH
    #files = os.listdir(root)
    
    #for f in files:
        #get_mentions(root+f)
    
    # conll/OntoNotes
    root = config.CONLL_PATH
    
    files = []
    for x in os.walk(root):
        for y in x[2]:
            #print(y)
            if y.endswith(".v4_gold_conll"):
                files.append(x[0]+"/"+y)
    
    for f in files:
        to_mmax(f, "/home/students/zimmermann/Projects/ncr/annotation/Entity/")
    
    
